Rice.require 'TableConverter'
Rice.require 'ExtensionLoader'
Rice.require 'Native'

-----------------------------------------------------------------------------
-- Local TABLE : INPUTTRAY COMMAND
-----------------------------------------------------------------------------
local INPUTTRAY_TABLE =	{
	autotrayselect		= 'ALL',
	tray1				= 'TRAY1',
	tray2				= 'TRAY2',
	tray3				= 'TRAY3',
	tray3lct			= 'LCT',
	tray4				= 'TRAY4',
	tray5				= 'TRAY5',
	tray6				= 'TRAY6',
	tray7				= 'TRAY7',
	bypasstray			= 'BYPASS',
	onesheetbypass		= 'TRAY7',
	frontbypasstray		= 'BYPASS',
	largecapacitytray	= 'LCT',
	interposer			= 'INSERTER',
	interposeruppertray	= 'INSERTER',
	interposerlowertray	= 'INSERTER2',
	bindinginterposer1	= 'BINDINSERTER',
	bindinginterposer2	= 'BINDINSERTER2'
}

-----------------------------------------------------------------------------
-- Local TABLE : STAPLE_TABLE COMMAND
-----------------------------------------------------------------------------
local STAPLE_TABLE = {
	off				= 'OFF',
	topleftslant	= 'LEFTTOPSLANTPORT',
	toprightslant	= 'RIGHTTOPSLANTPORT',
	topleft			= 'LEFTTOP',
	topright		= 'RIGHTTOP',
	['2attop']		= 'TOP2PORT',
	['2atleft']		= 'LEFT2PORT',
	['2atright']	= 'RIGHT2PORT',
	['2atcenter']	= 'BOOKLET'
}

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function isMember( value, tbl )
	for i = 1 , #tbl do
		if(tbl[i] == value) then
			return true
		end
	end
	return false
end

local function isEmpty( value )
	return ( not value or '' == value )
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function RoundOff( num, power_of_ten)
	if( not tonumber(num) or not tonumber(power_of_ten) ) then
		return nil
	end

	return tostring((math.floor(tonumber( num ) * (10 ^ ( -1 * tonumber(power_of_ten))) + 0.5)) * (10 ^ tonumber(power_of_ten)))
end

local function Multiply( a, b )
	if( not tonumber(a) or not tonumber(b) ) then
		return nil
	end

	return tostring(a * b)
end

local function GetMMValue( num )
	local MMValue = tonumber(RoundOff( Multiply( num, 0.1 ) , -1))
	if( not MMValue ) then
		return nil
	else
		return string.format( '%0.1f', MMValue )
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function isOrientationChanged(layout_value)
	return ( '2pagespersheet' == layout_value or '6pagespersheet' == layout_value )
end

function isPortrait(tbl)
	if( isOrientationChanged(tbl.layout) ) then
		return ( 'landscape' == tbl.orientation )
	else
		return ( 'landscape' ~= tbl.orientation )
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function isCollate(tbl)
	if( 'off' == tbl.enablecollate ) then
		return false
	else
		if( 'off' == tbl.ignoreapplicationcollate ) then
			return ( '0' ~= tbl.dmCollate )
		else
			return true
		end
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function ExpandTable( t )
	local retTable = {}

	local function Expand( t )
		if( 'table' == type(t) ) then
			for i, it in pairs( t ) do
				if( 'table' == type( it ) ) then
					Expand( it )
				elseif( 'string' == type( it ) ) then
					table.insert( retTable, it )
				elseif( 'function' == type( it ) ) then
					Expand( it() )
				end
			end
		else
			if( 'string' == type( t ) ) then
				table.insert( retTable, t )
			elseif( 'function' == type( t ) ) then
				Expand( t() )
			end
		end
	end

	Expand( t )
	return retTable
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function GetEncryptPassword( tbl )
	if( isMember( tbl.jobtype, { 'documentserver', 'documentserverrgb' } ) ) then
		if('3' == tbl.jobpasswordgeneration ) then
			return tbl.jt_documentserver_password3
		elseif('4' == tbl.jobpasswordgeneration ) then
			return tbl.jt_documentserver_password4
		else
			return tbl.jt_documentserver_password2
		end
	else
		if('3' == tbl.jobpasswordgeneration ) then
			return tbl.jt_jobpassword3
		elseif('4' == tbl.jobpasswordgeneration ) then
			return tbl.jt_jobpassword4
		else
			return tbl.jt_jobpassword2
		end
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function isInvalidJobCode( tbl )
	if( isMember( tbl.jobtype, { 'documentserver', 'documentserverrgb' } ) ) then
		return false
	elseif( isEmpty( tbl.userid ) or ( 'lockedprint' == tbl.jobtype and isEmpty( GetEncryptPassword( tbl ) ) ) ) then
		return true
	else
		return false
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function isFrontCoverWithBooklet( tbl )
	return ( 'on' == tbl.frontcover and '2atcenter' == tbl.staple and isMember( tbl.centerstaple_swapcover, {'1', '2', '3'} ) )
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local custom_media_table = {
	'CUSTOM1',
	'CUSTOM2',
	'CUSTOM3',
	'CUSTOM4',
	'CUSTOM5',
	'CUSTOM6',
	'CUSTOM7',
	'CUSTOM8'
}

local function isCustomMedia( tbl )
	if( 'on' == tbl.slipsheet ) then
		return isMember( tbl.jt_transparency_mediatype_jobcommand, custom_media_table )
	else
		return isMember( tbl.jt_standardpages_mediatype_jobcommand, custom_media_table )
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local ansiCodePage	= Native.Dll('kernel32.dll'):api('GetACP'):call_int()
local function isLeadByte( bytecode )
	local kernel32			= Native.Dll('kernel32.dll')

	return (kernel32:api('IsDBCSLeadByteEx')
	:with_int( ansiCodePage )
	:with_byte( bytecode )
	:call_int() ~= 0 )
end

local function GetTrimmedString( str, max_length )
	if (not str) or (not max_length) then
		return ''
	end

	local buf_size = math.min( string.len(str), max_length )

	local index = 1
	while index <= buf_size do
		if isLeadByte( string.byte( str, index ) ) then
			if index == buf_size then
				buf_size = buf_size - 1
				do break end
			end
			index = index + 2
		else
			if index == buf_size  then
				do break end
			end
			index = index + 1
		end
	end

	return string.sub( str, 1, buf_size )
end

local function isWrongString( str, max_length )
	if (not str) or (not max_length) then
		return true
	end

	if string.len(str) > max_length then
		return true
	else
		return false
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function GetJobName( tbl )
	if( isEmpty(tbl.jobname) ) then
		return '""'
	else
		return '"' .. GetTrimmedString( tbl.jobname, 232 ) .. '"'
	end
end

local function GetJob_Name( tbl )
	if( isEmpty(tbl.jobname) ) then
		return '""'
	else
		return '"' .. GetTrimmedString( tbl.jobname, 80 ) .. '"'
	end
end

-----------------------------------------------------------------------------
-- Conversion Table : setting_table -> job_command_table
-----------------------------------------------------------------------------
local convert_table = {

	userid = {
		item = 'USERID',
		values = function(value, tbl)
			if( isEmpty( value ) or isWrongString( value, 233 ) ) then
				return '"?"'
			else
				return  '"' .. value .. '"'
			end
		end
	},

	jobtype = {
		item = function(value, tbl)
			local password = GetEncryptPassword( tbl )

			local function PasswordItem()
				if( isEmpty( password ) or ( '4' == tbl.jobpasswordgeneration and 'on' == tbl.userauthentication ) ) then
					return nil
				end

				if( '4' == tbl.jobpasswordgeneration ) then
					return {'JOBPASSWORD' .. tbl.jobpasswordgeneration, 'JOBPASSWORDENCRYPTMETHOD', 'JOBPASSWORDENCRYPTSETTING'}
				elseif( true == isMember( tbl.jobpasswordgeneration, { '2', '3' } ) ) then
					return 'JOBPASSWORD' .. tbl.jobpasswordgeneration
				else
					return 'JOBPASSWORD'
				end
			end

			local function JobIDItem()
				local fname = (function ()
					if( isMember( value, { 'documentserver', 'documentserverrgb' } ) ) then
						return tbl.documentserver_filename
					else
						return tbl.filename
					end
				end)()

				if ( isEmpty( fname ) ) then
					return nil
				else
					return 'JOBID'
				end
			end

			local function OwnerIDItem()
				if ( isEmpty( tbl.username ) ) then
					return nil
				else
					return 'OWNERID'
				end
			end

			if ( isInvalidJobCode( tbl ) ) then
				return nil --NormalPrint̎s
			end

			local t = {
				normalprint			= nil,
				sampleprint			= nil,
				lockedprint			= { nil			, nil			, nil		, PasswordItem	, nil			},
				holdprint			= { 'SAVEMODE'	, nil			, JobIDItem	, nil			, nil			},
				storedprint			= { 'SAVEMODE'	, 'ACCESSMODE'	, JobIDItem	, PasswordItem	, nil			},
				storedprintshared	= { 'SAVEMODE'	, 'ACCESSMODE'	, JobIDItem	, PasswordItem	, nil			},
				storeandprint		= { 'SAVEMODE'	, 'ACCESSMODE'	, JobIDItem	, PasswordItem	, nil			},
				storeandprintshared	= { 'SAVEMODE'	, 'ACCESSMODE'	, JobIDItem	, PasswordItem	, nil			},
				documentserver		= { 'DISKIMAGE'	, nil			, JobIDItem	, PasswordItem	, OwnerIDItem	},
				documentserverrgb	= { nil			, nil			, JobIDItem	, PasswordItem	, OwnerIDItem	}
			}

			return t[value] and ExpandTable( t[value] ) or ExpandTable( t['normalprint'] )
		end,
		values = function( value, tbl )
			local password = GetEncryptPassword( tbl )

			local function JobIDValue()
				local fname = (function ()
					if( isMember( value, { 'documentserver', 'documentserverrgb' } ) ) then
						return tbl.documentserver_filename
					else
						return tbl.filename
					end
				end)()

				if ( isEmpty( fname ) or isWrongString( fname, 20 ) ) then
					return nil
				else
					return '"' .. fname .. '"'
				end
			end

			local function OwnerIDValue()
				if ( isEmpty( tbl.username ) or isWrongString( tbl.username, 20 ) ) then
					return nil
				else
					return '"' .. tbl.username .. '"'
				end
			end

			if ( isInvalidJobCode( tbl ) ) then
				return nil --NormalPrint̎s
			end

			local function PasswordValue()
				if( isEmpty( password ) or ( '4' == tbl.jobpasswordgeneration and 'on' == tbl.userauthentication ) ) then
					return nil
				end

				if( '4' == tbl.jobpasswordgeneration ) then
					if( isWrongString( password, 227 ) ) then
						return nil
					else
						return { '"' .. password .. '"', '0', '2' }
					end
				elseif( '3' == tbl.jobpasswordgeneration ) then
					if( isWrongString( password, 227 ) ) then
						return nil
					else
						return '"' .. password .. '"'
					end
				elseif( '2' == tbl.jobpasswordgeneration ) then
					if( isWrongString( password, 8 ) ) then
						return nil
					else
						return '"' .. password .. '"'
					end
				else
					if( isWrongString( password, 4 ) ) then
						return nil
					else
						return '"' .. password .. '"'
					end
				end
			end

			local t = {
				normalprint			= nil,
				sampleprint			= nil,
				lockedprint			= { nil		, nil		, nil			, PasswordValue	, nil			},
				holdprint			= { 'OFF'	, nil		, JobIDValue	, nil			, nil			},
				storedprint			= { 'ON'	, 'PRIVATE'	, JobIDValue	, PasswordValue	, nil			},
				storedprintshared	= { 'ON'	, 'PUBLIC'	, JobIDValue	, PasswordValue	, nil			},
				storeandprint		= { 'ON'	, 'PRIVATE'	, JobIDValue	, PasswordValue	, nil			},
				storeandprintshared	= { 'ON'	, 'PUBLIC'	, JobIDValue	, PasswordValue	, nil			},
				documentserver		= { 'ON'	, nil		, JobIDValue	, PasswordValue	, OwnerIDValue	},
				documentserverrgb	= { nil		, nil		, JobIDValue	, PasswordValue	, OwnerIDValue	}
			}

			return t[value] and ExpandTable( t[value] ) or ExpandTable( t['normalprint'] )
		end
	},

	faceup = {
		item	= {'PAPERFACE'},
		values	= function( value, tbl )
			local function DownOrUp()
				if( 'on' == tbl.multisheetfold and 'off' ~= tbl.duplex ) then
					return 'DOWN'
				else
					return 'UP'
				end
			end

			local retVal = ExpandTable(({
				off								= { nil			},
				rightzfold						= { nil			},
				leftzfold						= { nil			},
				bottomzfold						= { nil			},
				insidehalffold					= { 'DOWN'		},
				outsidehalffold					= { DownOrUp	},
				rightortopinsidespiralfold		= { 'DOWN'		},
				leftorbottominsidespiralfold	= { 'DOWN'		},
				rightortopoutsidespiralfold		= { DownOrUp	},
				leftorbottomoutsidespiralfold	= { DownOrUp	},
				outsidezigzagfold				= { nil			},
				insidezigzagfold				= { nil			},
				rightortopinsidedoublefold		= { 'DOWN'		},
				leftorbottominsidedoublefold	= { 'DOWN'		},
				rightortopoutsidedoublefold		= { 'UP'		},
				leftorbottomoutsidedoublefold	= { 'UP'		},
				insidedoublegatefold			= { 'DOWN'		},
				outsidedoublegatefold			= { 'UP'		}
			})[tbl.fold])[1]

			if( not retVal ) then
				if( 'on' == value ) then
					return 'UP'
				else
					return 'DOWN'
				end
			else
				return retVal
			end
		end
	},

	fold = {
		item = {
			off								=  { 'FOLD' },
			rightzfold						=  { 'FOLD' },
			leftzfold						=  { 'FOLD' },
			bottomzfold						=  { 'FOLD' },
			insidehalffold					=  { 'FOLD', 'FOLDOPENSIDE' },
			outsidehalffold					=  { 'FOLD', 'FOLDOPENSIDE' },
			rightortopinsidespiralfold		=  { 'FOLD', 'FOLDOPENSIDE' },
			leftorbottominsidespiralfold	=  { 'FOLD', 'FOLDOPENSIDE' },
			rightortopoutsidespiralfold		=  { 'FOLD', 'FOLDOPENSIDE' },
			leftorbottomoutsidespiralfold	=  { 'FOLD', 'FOLDOPENSIDE' },
			outsidezigzagfold				=  { 'FOLD', 'FOLDOPENSIDE' },
			insidezigzagfold				=  { 'FOLD', 'FOLDOPENSIDE' },
			rightortopinsidedoublefold		=  { 'FOLD', 'FOLDOPENSIDE' },
			leftorbottominsidedoublefold	=  { 'FOLD', 'FOLDOPENSIDE' },
			rightortopoutsidedoublefold		=  { 'FOLD', 'FOLDOPENSIDE' },
			leftorbottomoutsidedoublefold	=  { 'FOLD', 'FOLDOPENSIDE' },
			insidedoublegatefold			=  { 'FOLD', 'FOLDOPENSIDE' },
			outsidedoublegatefold			=  { 'FOLD', 'FOLDOPENSIDE' },
		},
		values = function( value, tbl )
			local is_portrait = isPortrait( tbl )

			local function TopOrRight()
				if( is_portrait ) then
					return 'TOP'
				else
					return 'RIGHT'
				end
			end

			local function BottomOrLeft()
				if( is_portrait ) then
					return 'BOTTOM'
				else
					return 'LEFT'
				end
			end

			local t = {
				off								= { 'OFF'		, nil			},
				rightzfold						= { 'ZRIGHT'	, nil			},
				leftzfold						= { 'ZLEFT'		, nil			},
				bottomzfold						= { 'ZBOTTOM'	, nil			},
				insidehalffold					= { 'TWO'		, TopOrRight	},
				outsidehalffold					= { 'TWO'		, BottomOrLeft	},
				rightortopinsidespiralfold		= { 'INSIDE3'	, TopOrRight	},
				leftorbottominsidespiralfold	= { 'INSIDE3'	, BottomOrLeft	},
				rightortopoutsidespiralfold		= { 'INSIDE3'	, TopOrRight	},
				leftorbottomoutsidespiralfold	= { 'INSIDE3'	, BottomOrLeft	},
				outsidezigzagfold				= { 'OUTSIDE3'	, TopOrRight	},
				insidezigzagfold				= { 'OUTSIDE3'	, BottomOrLeft	},
				rightortopinsidedoublefold		= { 'SIMPLE4'	, TopOrRight	},
				leftorbottominsidedoublefold	= { 'SIMPLE4'	, BottomOrLeft	},
				rightortopoutsidedoublefold		= { 'SIMPLE4'	, TopOrRight	},
				leftorbottomoutsidedoublefold	= { 'SIMPLE4'	, BottomOrLeft	},
				insidedoublegatefold			= { 'GATEFOLD4'	, TopOrRight	},
				outsidedoublegatefold			= { 'GATEFOLD4'	, BottomOrLeft	}
			}

			return ExpandTable( t[value] )
		end,
		default = { 'FOLD', 'OFF' }
	},

	orientation = {
		item = 'ORIENTATION',
		values = function( value, tbl )
			if( not isPortrait( tbl ) ) then
				return 'LANDSCAPE'
			else
				return 'PORTRAIT'
			end
		end
	},

	duplex = {
		item = { 'DUPLEX', 'BINDING' },
		values = function( value, tbl )
			if( tbl.booklet and 'off' ~= tbl.booklet ) then
				return nil
			else
				if( 'off' == value ) then
					return 'OFF'
				else
					if( isOrientationChanged(tbl.layout) ) then
						if( 'opentoleft' == value ) then
							if( 'portrait' == tbl.orientation ) then
								return {'ON', 'SHORTEDGE'}
							else
								return {'ON', 'LONGEDGE'}
							end
						else
							if( 'portrait' == tbl.orientation ) then
								return {'ON', 'LONGEDGE'}
							else
								return {'ON', 'SHORTEDGE'}
							end
						end
					else
						if( 'opentoleft' == value ) then
							if( 'portrait' == tbl.orientation ) then
								return {'ON', 'LONGEDGE'}
							else
								return {'ON', 'SHORTEDGE'}
							end
						else
							if( 'portrait' == tbl.orientation ) then
								return {'ON', 'SHORTEDGE'}
							else
								return {'ON', 'LONGEDGE'}
							end
						end
					end
				end
			end
		end
	},

	resolution = {
		item = function(value, tbl)
			if( '1200x600' ~= value and tbl.gradation and isMember( tbl.gradation, { 'fast', 'standard', 'fine' } ) ) then
				return 'RESOLUTION'
			else
				return { 'RESOLUTION', 'BITSPERDOT' }
			end
		end,
		values = function(value, tbl)
			if( '1200x600' ~= value and tbl.gradation and isMember( tbl.gradation, { 'fast', 'standard', 'fine' } ) ) then
				return {
					['1200x600']	= '600',
					['1200'] 		= '1200',
					['600'] 		= '600'
				}
			else
				return {
					['1200x600']	= { '600'	, '2' },
					['1200'] 		= { '1200'	, '1' },
					['600'] 		= { '600'	, '1' }
				}
			end
		end
	},

	gradation = {
		item = 'BITSPERDOT',
		values = {
			fast		= '1',
			standard	= '2',
			fine		= '4'
		}
	},

	copies = {
		item = function( value, tbl )
			if( isCollate(tbl) or isMember( tbl.staple, { 'topleftslant', 'toprightslant', 'topleft', '2atright', '3atleft', '4atleft', 'multiatleft', 'multiatright', 'multiattop' } ) ) then
				return {'COPIES', 'QTY' }
			else
				return 'COPIES'
			end
		end,
		values = function( value, tbl )
			if( isCollate(tbl) or isMember( tbl.staple, { 'topleftslant', 'toprightslant', 'topleft', '2atright', '3atleft', '4atleft', 'multiatleft', 'multiatright', 'multiattop' } ) ) then
				return {'1', value }
			else
				return value
			end
		end
	},

	colorbw = {
		item		= { 'DATAMODE', 'RENDERMODE' },
		values	= {
			color			= { 'COLOR'		, 'COLOR'		},
			blackandwhite	= { 'GRAYSCALE'	, 'GRAYSCALE'	}
		},
		default = { { 'DATAMODE', 'RENDERMODE' }, { 'COLOR' , 'COLOR' } }
	},

	jt_paper_jobcommand = {
		item = 'PAPER',
		values = function(value) return value end
	},

	jt_standardpages_mediatype_jobcommand = {
		item = function( value, tbl )
			if( 'on' == tbl.slipsheet ) then
				return nil
			else
				return 'MEDIATYPE'
			end
		end,
		values = function( value, tbl )
			if( 'on' == tbl.slipsheet ) then
				return nil
			else
				return value
			end
		end
	},

	jt_transparency_mediatype_jobcommand = {
		item = 'MEDIATYPE',
		values = function( value, tbl )
			if( 'on' == tbl.slipsheet ) then
				return value
			else
				return nil
			end
		end
	},

	booklet = {
		item = {'DUPLEX', 'BINDING'},
		values = function( value, tbl )
			if( 'off' == value ) then
				return nil
			else
				return {'ON', 'SHORTEDGE'}
			end
		end
	},

	insertseparationsheet = {
		item = function( value, tbl )
			if( 'on' == value ) then
				return {'SEPARATIONSHEETMODE', 'SEPARATIONSHEETCYCLE'}
			else
				return 'SEPARATIONSHEETMODE'
			end
		end,
		values = function( value, tbl )
			if( 'on' == value ) then
				return { 'SORT', tbl.insertseparationsheet_percopies }
			else
				return 'OFF'
			end
		end
	},

	edgetoedgeprint = {
		item = 'EDGETOEDGE',
		values = {
			on	= 'YES',
			off	= 'NO'
		},
		default = { 'EDGETOEDGE', 'NO' }
	},

	edgesmoothing = {
		item = 'SMOOTHING',
		values = {
			on	= 'ON',
			off	= 'OFF'
		},
		default = { 'SMOOTHING', 'OFF' }
	},

	tonersaving = {
		item = function( value, tbl )
			if( nil == tbl.economycolormode ) then
				return 'ECONOMODE'
			else
				return nil
			end
		end,
		values = function( value, tbl )
			if( 'on' == value ) then
				return 'ON'
			else
				return 'OFF'
			end
		end
	},

	economycolormode = {
		item = 'ECONOMODE',
		values = {
			on	= 'ON',
			off	= 'OFF'
		},
		default = { 'ECONOMODE', 'OFF' }
	},

	blackoverprint = {
		item = 'BLACKOVERPRINT',
		values = function(value, tbl)
			if( '1' ~= tbl.boptype ) then
				return nil
			else
				if( 'manual' == tbl.coloroption ) then
					if( 'on' == value ) then
						return 'ON'
					else
						return 'OFF'
					end
				else
					return 'OFF'
				end
			end
		end
	},

	staple = {
		item = {'STAPLE'},
		values = STAPLE_TABLE,
		default = { 'STAPLE', 'OFF' }
	},

	multisheetfold = {
		item = 'PILINGFOLD',
		values = function( value, tbl )
			if( 'on' == value ) then
				return 'ON'
			else
				return 'OFF'
			end
		end,
	},

	ringbinding = {
		item = {
			off							= 'RINGBINDMODE',
			ringbindingleft				= { 'RINGBINDMODE'	, 'BINDPOSITION' },
			ringbindingtop				= { 'RINGBINDMODE'	, 'BINDPOSITION' },
			ringbindingright			= { 'RINGBINDMODE'	, 'BINDPOSITION' },
			ringbindingleftpunchonly	= { 'RINGBINDMODE'	, 'BINDPOSITION' },
			ringbindingtoppunchonly		= { 'RINGBINDMODE'	, 'BINDPOSITION' },
			ringbindingrightpunchonly	= { 'RINGBINDMODE'	, 'BINDPOSITION' }
		},
		values = {
			off							= 'OFF',
			ringbindingleft				= { 'RING'	, 'LEFTPORT'	},
			ringbindingtop				= { 'RING'	, 'TOPLAND'		},
			ringbindingright			= { 'RING'	, 'RIGHTPORT'	},
			ringbindingleftpunchonly	= { 'PUNCH'	, 'LEFTPORT'	},
			ringbindingtoppunchonly		= { 'PUNCH'	, 'TOPLAND'		},
			ringbindingrightpunchonly	= { 'PUNCH'	, 'RIGHTPORT'	}
		},
		default = { 'RINGBINDMODE', 'OFF' }
	},

	perfectbinding = {
		item = 'PERFECTBIND',
		values = {
			on	= 'ON',
			off	= 'OFF'
		},
		default = { 'PERFECTBIND', 'OFF' }
	},

	perfectbinding_direction = {
		item = 'BINDPOSITION',
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				local t ={
					leftbinding		= 'LEFTPORT',
					rightbinding	= 'RIGHTPORT',
					topbinding		= 'TOPLAND'
				}
				return t[value] and t[value] or t['leftbinding']
			else
				return nil
			end
		end
	},

	usercode = {
		item = 'USERCODE',
		values = function(value, tbl)
			if( '2' == tbl.controllerplatformtype) then
				return nil
			else
				if( isWrongString( tbl.usercode, 231 ) ) then
					return nil
				else
					return '"' .. tbl.usercode .. '"'
				end
			end
		end
	},

	jobname = {
		item = 'JOBNAME',
		values = function(value, tbl)
			return GetJobName( tbl )
		end
	},

	timeinstant = {
		item = 'TIME',
		values = function(value, tbl)
			if( isEmpty( tbl.timeinstant ) ) then
				return '"00:00:00"'
			else
				return '"' .. GetTrimmedString( tbl.timeinstant, 8 ) .. '"'
			end
		end
	},

	date = {
		item = 'DATE',
		values = function(value, tbl)
			if( isEmpty( tbl.date ) ) then
				return '"2000/01/01"'
			end
			return '"' .. GetTrimmedString( tbl.date, 10 ) .. '"'
		end
	},

	jt_trackid = {
		item = 'TRACKID',
		values = function(value, tbl)
			return '"' .. GetTrimmedString( tbl.jt_trackid, 20 ) .. '"'
		end
	},

	windowslogonusername = {
		item = 'HOSTLOGINNAME',
		values = function(value, tbl)
			return '"' .. GetTrimmedString( tbl.windowslogonusername, 20 ) .. '"'
		end
	},

	billingcode = {
		item = 'BILLINGCODE',
		values = function(value, tbl)
			if( isMember( tbl.jobtype, { 'documentserver', 'documentserverrgb' } ) ) then
				return nil
			else
				if isWrongString( tbl.billingcode, 228 ) then
					return nil
				else
					return '"' .. tbl.billingcode .. '"'
				end
			end
		end
	},

	standardpages_inputtray = {
		item = function( value, tbl )
			if( 'on' == tbl.slipsheet ) then
				return nil
			else
				if( 'on' == tbl.perfectbinding ) then
					return { 'AUTOTRAYCHANGE', 'PERFECTBINDPAPERTRAY' }
				else
					if( 'on' == tbl.frontcover or 'on' == tbl.backcover or 'on' == tbl.chapterpages ) then
						return { 'AUTOTRAYCHANGE', 'TRAY' }
					else
						return 'AUTOTRAYCHANGE'
					end
				end
			end
		end,
		values = {
			autotrayselect		= {'ON','ALL'},
			tray1				= {'OFF','TRAY1'},
			tray2				= {'OFF','TRAY2'},
			tray3				= {'OFF','TRAY3'},
			tray3lct			= {'OFF','LCT'},
			tray4				= {'OFF','TRAY4'},
			tray5				= {'OFF','TRAY5'},
			tray6				= {'OFF','TRAY6'},
			tray7				= {'OFF','TRAY7'},
			bypasstray			= {'OFF','BYPASS'},
			onesheetbypass		= {'OFF','TRAY7'},
			frontbypasstray		= {'OFF','BYPASS'},
			largecapacitytray	= {'OFF','LCT'}
		}
	},

	transparency_inputtray = {
		item = function( value, tbl )
			if( 'on' == tbl.slipsheet ) then
				return { 'AUTOTRAYCHANGE', 'TRAY' }
			else
				return nil
			end
		end,
		values = {
			tray1				= {'OFF','TRAY1'},
			tray2				= {'OFF','TRAY2'},
			tray3				= {'OFF','TRAY3'},
			tray3lct			= {'OFF','LCT'},
			tray4				= {'OFF','TRAY4'},
			tray5				= {'OFF','TRAY5'},
			tray6				= {'OFF','TRAY6'},
			tray7				= {'OFF','TRAY7'},
			bypasstray			= {'OFF','BYPASS'},
			onesheetbypass		= {'OFF','TRAY7'},
			frontbypasstray		= {'OFF','BYPASS'},
			largecapacitytray	= {'OFF','LCT'}
		}
	},

	standardpages_tabstockprinting = {
		item = 'TABPOSITION',
		values = {
			off	= 'OFF',
			on	= 'RIGHT'
		}
	},

	standardpages_shiftprintposition = {
		item = function(value,tbl)
			if( 'on' == tbl.standardpages_tabstockprinting ) then
				return 'TABSTOCKSHIFTPRINT'
			else
				return nil
			end
		end,
		values = {
			off	= 'OFF',
			on	= 'ON'
		}
	},

	frontcover =	{
		item = 'FRONTCOVERPRINT',
		values = function(value, tbl)
			if( 'off' == value or ( isFrontCoverWithBooklet( tbl ) and isMember( tbl.centerstaple_swapcover, {'2','3'} ) ) ) then
				return 'OFF'
			else
				return nil
			end
		end
	},

	frontcover_printmode =	{
		item = function(value,tbl)
			if( isFrontCoverWithBooklet( tbl ) ) then
				return 'BACKCOVERPRINT'
			elseif( 'on' == tbl.frontcover ) then
				return  'FRONTCOVERPRINT'
			else
				return nil
			end
		end,
		values = function(value, tbl)
			if( isFrontCoverWithBooklet( tbl ) ) then
				if( '3' == tbl.centerstaple_swapcover ) then
					return {
						leaveblank	= 	'BLANK',
						oneside		= 	'INNERCOPY',
						bothsides	= 	'INNERCOPY'
					}
				elseif( isMember( tbl.centerstaple_swapcover, {'1','2'} ) ) then
					return {
						leaveblank	= 	'BLANK',
						oneside		= 	'OUTERCOPY',
						bothsides	= 	'OUTERCOPY'
					}
				end
			elseif( 'on' == tbl.frontcover ) then
				return {
					leaveblank	= 	'BLANK',
					oneside		= 	'SIMPLEXCOPY',
					bothsides	= 	'DUPLEXCOPY'
				}
			else
				return nil
			end
		end
	},

	frontcover_inputtray =	{
		item = function( value, tbl )
			if( isFrontCoverWithBooklet( tbl ) and '1' == tbl.centerstaple_swapcover ) then
				return 'BACKCOVERPRINTTRAY'
			elseif( 'on' == tbl.frontcover ) then
				return 'FRONTCOVERPRINTTRAY'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	backcover =	{
		item = 'BACKCOVERPRINT',
		values = function(value, tbl)
			if( 'off' == value and not isFrontCoverWithBooklet(tbl) ) then
				return 'OFF'
			else
				return nil
			end
		end
	},

	backcover_printmode =	{
		item = function(value,tbl)
			if( 'on' == tbl.backcover and not isFrontCoverWithBooklet(tbl) ) then
				return  'BACKCOVERPRINT'
			else
				return nil
			end
		end,
		values = {
			leaveblank			= 'BLANK',
			printonbackcover 	= 'INNERCOPY',
			printonouter 		= 'OUTERCOPY'
		}
	},

	backcover_inputtray =	{
		item = function( value, tbl )
			if( 'on' == tbl.backcover and not isFrontCoverWithBooklet(tbl) ) then
				return 'BACKCOVERPRINTTRAY'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	slipsheet = {
		item = 'SLIPSHEETPRINT',
		values = function( value, tbl )
			if( 'off' == value ) then
				return 'OFF'
			else
				if( 'printonslipsheet' == tbl.slipsheet_printmode ) then
					return 'COPY'
				else
					return 'BLANK'
				end
			end
		end
	},

	slipsheet_inputtray =	{
		item = function( value, tbl )
			if( 'on' == tbl.slipsheet ) then
				return 'SLIPSHEETPRINTTRAY'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	chapterpages = {
		item =	'CHAPTERPRINT',
		values ={
			on	= 'ON'
		}

	},

	jt_chaptertray1 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY1'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jt_chaptermode1 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX1'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jt_chaptertray2 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY2'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jt_chaptermode2 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX2'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jt_chaptertray3 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY3'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jt_chaptermode3 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX3'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jt_chaptertray4 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY4'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jt_chaptermode4 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX4'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jt_chaptertray5 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY5'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jt_chaptermode5 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX5'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jt_chaptertray6 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY6'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jt_chaptermode6 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX6'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jt_chaptertray7 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY7'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jt_chaptermode7 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX7'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jt_chaptertray8 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY8'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jt_chaptermode8 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX8'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jt_charset = {
		item =  { 'HOSTCHARSET2', 'DISPCHARSET2' },
		values = function(value, tbl)
			if( isMember( value, { '1', '2', '3', '5', '6', '8', '9', '13' } )) then
				return { value, value }
			else
				return { '2', '2' }
			end
		end
	},

	loginusername = {
		item = function( value, tbl )
			if( 'on' == tbl.userauthentication and not isEmpty( value ) and not isWrongString( value, 217 ) ) then
				return { 'AUTHENTICATIONUSERNAME', 'AUTHENTICATIONUSERNAMECHARSET', 'AUTHENTICATIONPASSWORDCHARSET' }
			else
				return { 'AUTHENTICATIONUSERNAME', 'AUTHENTICATIONUSERNAMECHARSET'}
			end
		end,
		values = function(value, tbl)
			if( 'on' == tbl.userauthentication and not isEmpty( value ) and not isWrongString( value, 217 ) ) then
				if( isMember( tbl.jt_charset, { '1', '2', '3', '5', '6', '8', '9', '13' } )) then
					return { '"' .. value .. '"', tbl.jt_charset, tbl.jt_charset }
				else
					return { '"' .. value .. '"', '2', '2' }
				end
			else
				if( isMember( tbl.jt_charset, { '1', '2', '3', '5', '6', '8', '9', '13' } )) then
					return { '""', tbl.jt_charset }
				else
					return { '""', '2' }
				end
			end
		end
	},

	withencryption = {
		item ={ 'AUTHENTICATIONENCRYPTSETTING', 'AUTHENTICATIONENCRYPTMETHOD' },
		values = function(value, tbl)
			if( 'on' == tbl.userauthentication  and not isEmpty( tbl.loginusername )) then
				if( 'on' == value ) then
					return { '2', '0' }
				else
					return { '1', '0' }
				end
			else
				return nil
			end
		end
	},

	jt_loginpassword3 = {
		item = 'AUTHENTICATIONPASSWORD',
		values = function(value, tbl)
			if( 'on' == tbl.userauthentication ) then
				if( isEmpty( tbl.loginusername ) or isWrongString( value, 217 ) ) then
					return nil
				else
					return '"' .. value .. '"'
				end
			else
				return nil
			end
		end
	},

	jt_loginpassword4 = {
		item = 'AUTHENTICATIONPASSWORD',
		values = function(value, tbl)
			if( 'on' == tbl.userauthentication ) then
				if( isEmpty( tbl.loginusername ) or isWrongString( value, 217 ) ) then
					return nil
				else
					return '"' .. value .. '"'
				end
			else
				return nil
			end
		end
	},

	joboffset = {
		item = 'JOBOFFSET',
		values = function(value, tbl)
			if( isCollate(tbl) ) then
				return {
						off		= 'OFF',
						rotate	= 'ROTATE',
						shift	= 'SHIFT'
				}
			else
				return 'OFF'
			end
		end,
		default = { 'JOBOFFSET', 'OFF' }
	},

	outputtray = {
		item = 'OUTBIN',
		values = {
			printerdefault				= nil,
			standardtray				= 'UPPER',
			copytray					= 'UPPER',
			internaltray				= 'UPPER',
			externaltray				= 'UPPER',
			finishertray1				= 'FINISHERPROOF',
			finishertray2				= 'FINISHERSHIFT',
			finisheruppertray			= 'FINISHERPROOF',
			finishershifttray			= 'FINISHERSHIFT',
			finishershifttray1			= 'FINISHERPROOF',
			finishershifttray2			= 'FINISHERSHIFT',
			bookletprocessortray		= 'FINISHERSHIFT',
			finisherbooklettray			= 'FINISHERBOOKLET',
			lefttray1					= 'LOWER2',
			lefttray2					= 'LOWER3',
			mailboxprooftray			= 'OPTIONALOUTPUTBIN1',	-- T.B.D
			mailboxtray1				= 'OPTIONALOUTPUTBIN2',
			mailboxtray2				= 'OPTIONALOUTPUTBIN3',
			mailboxtray3				= 'OPTIONALOUTPUTBIN4',
			mailboxtray4				= 'OPTIONALOUTPUTBIN5',
			mailboxtray5				= 'OPTIONALOUTPUTBIN6',
			mailboxtray6				= 'OPTIONALOUTPUTBIN7',
			mailboxtray7				= 'OPTIONALOUTPUTBIN8',
			mailboxtray8				= 'OPTIONALOUTPUTBIN9',
			mailboxtray9				= 'OPTIONALOUTPUTBIN10',
			onesheetoutputtray			= 'LOWER',
			slidesorttray				= 'UPPER',
			internaltray1				= 'UPPER',
			internaltray2				= 'INNER',
			internalshifttray			= 'UPPER',
			righttray					= 'LOWER',
			loadingtray					= 'LOADINGTRAY',
			ringstacker					= 'RINGSTACKER',
			largestackertray			= 'STACKER',
			largestackertray2			= 'STACKER2',
			stackerprooftray			= 'STACKERPROOF',
			foldprooftray				= 'FOLDPROOF',
			lefttray					= 'LOWER'
		},
	},

	perfectbinding_imgpos_along_mm = {
		item		= { 'IMAGEVERTADJUSTMENT' },
		values	= function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				return GetMMValue( value )
			else
				return nil
			end
		end
	},

	perfectbinding_imgpos_across_mm = {
		item		= { 'IMAGEHORIZADJUSTMENT' },
		values	= function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				return GetMMValue( value )
			else
				return nil
			end
		end
	},

	 perfectbinding_customsize_width_mm = {
		item		= { 'PERFECTBINDPAPERCUSTOMWIDTH' },
		values	= function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				return GetMMValue( value )
			else
				return nil
			end
		end
	},

	perfectbinding_customsize_length_mm = {
		item		= { 'PERFECTBINDPAPERCUSTOMLENGTH' },
		values	= function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				return GetMMValue( value )
			else
				return nil
			end
		end
	},

	jt_prefectbind_paper_jobcommand = {
		item = 'PERFECTBINDPAPERSIZE',
		values =  function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				return value
			else
				return nil
			end
		end
	},

	punch = {
		item = 'PUNCH',
		values =  {
			['off']				= 'OFF',
			['2atleft']			= 'LEFTPORT',
			['2attop']			= 'TOPPORT',
			['2atright']		= 'RIGHTPORT',
			['3atleft']			= 'LEFTPORT',
			['3attop']			= 'TOPPORT',
			['3atright']		= 'RIGHTPORT',
			['4atleft']			= 'LEFTPORT',
			['4attop']			= 'TOPPORT',
			['4atright']		= 'RIGHTPORT',
			['multiatleft']		= 'LEFTPORT',
			['multiatright']	= 'RIGHTPORT',
			['multiattop']		= 'TOPPORT'
		},
		default = { 'PUNCH', 'OFF' }
	},

	punchunit = {
		item = 'PUNCHHOLE',
		values = function(value, tbl)
			local punch_value = tbl.punch

			return {
				['none']				= nil,
				['2holestypejp']		=	({
													['2atleft']			= 'JP2',
													['2attop']			= 'JP2',
													['2atright']		= 'JP2',
												})[punch_value],
				['2holestypeeu']		=	({
													['2atleft']			= 'JP2',
													['2attop']			= 'JP2',
													['2atright']		= 'JP2',
												})[punch_value],
				['2holestypeus2']		=	({
													['2atleft']			= 'US2',
													['2attop']			= 'US2',
													['2atright']		= 'US2',
												})[punch_value],
				['2and3holestypejp']	=	({
													['2atleft']			= 'JP2',
													['2attop']			= 'JP2',
													['2atright']		= 'JP2',
													['3atleft']			= 'US3',
													['3attop']			= 'US3',
													['3atright']		= 'US3',
												})[punch_value],
				['2and3holestypeus']	=	({
													['2atleft']			= 'US2',
													['2attop']			= 'US2',
													['2atright']		= 'US2',
													['3atleft']			= 'US3',
													['3attop']			= 'US3',
													['3atright']		= 'US3',
												})[punch_value],
				['2and4holestypeeu']	=	({
													['2atleft']			= 'JP2',
													['2attop']			= 'JP2',
													['2atright']		= 'JP2',
													['4atleft']			= 'EU4',
													['4attop']			= 'EU4',
													['4atright']		= 'EU4',
												})[punch_value],
				['3holestypeus3']		=	({
													['3atleft']			= 'US3',
													['3attop']			= 'US3',
													['3atright']		= 'US3',
												})[punch_value],
				['4holestypene']		=	({
													['4atleft']			= 'NEU4',
													['4attop']			= 'NEU4',
													['4atright']		= 'NEU4',
												})[punch_value],
				['4holestypeeu']		=	({
													['4atleft']			= 'EU4',
													['4attop']			= 'EU4',
													['4atright']		= 'EU4',
												})[punch_value]
			}
		end
	},

	acc_punch_multihole = {
		item = 'PUNCHHOLE',
		values = function(value, tbl)
			local punch_value = tbl.punch

			return {
				['on']				=	({
													['multiatleft']     = 'MULTI',
													['multiatright']    = 'MULTI',
													['multiattop']      = 'MULTI'
												})[punch_value],
				['off']				=	nil
			}
		end
	},

	perfectbinding_cuttype = {
		item		= 'TRIMMODE',
		values	= function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				local t = {
					donotcut						= 'OFF',
					cutforeedgespecifyfinishingsize	= 'SINGLECUTFINISHSIZE',
					cutforeedge						= 'SINGLECUT',
					cut3edgesspecifyfinishingsize	= 'TRIPLECUTFINISHSIZE',
					cut3edges						= 'TRIPLECUT'
				}

				return t[value] and t[value] or t['donotcut']
			else
				return nil
			end
		end
	},

	jt_prefectbind_cutpos_paper_jobcommand = {
		item = { 'TRIPLECUTFINISHSIZE' },
		values =  function(value, tbl)
			if( 'on' == tbl.perfectbinding and 'cut3edgesspecifyfinishingsize' == tbl.perfectbinding_cuttype) then
				return value
			else
				return nil
			end
		end
	},

	perfectbinding_cutpos_head_mm = {
		item = 'TOPCUT',
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding and  'cut3edges' == tbl.perfectbinding_cuttype) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	perfectbinding_cutpos_tail_mm = {
		item = 'BOTTOMCUT',
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding  and 'cut3edges' == tbl.perfectbinding_cuttype) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	perfectbinding_cutpos_fore_mm = {
		item = 'SIDEEDGECUT',
		values = function(value, tbl)
			if( 'on' ~= tbl.perfectbinding ) then
				return nil
			elseif('cutforeedge' == tbl.perfectbinding_cuttype or 'cut3edges' == tbl.perfectbinding_cuttype ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	perfectbinding_cutpos_width_mm = {
		item = 'TRIPLECUTFINISHSIZECUSTOMWIDTH',
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding and   'cut3edgesspecifyfinishingsize' == tbl.perfectbinding_cuttype and 'CUSTOM' == tbl.jt_prefectbind_cutpos_paper_jobcommand ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	perfectbinding_cutpos_length_mm = {
		item = function(value, tbl)
			if( 'on' ~= tbl.perfectbinding ) then
				return nil
			elseif( 'cut3edgesspecifyfinishingsize' == tbl.perfectbinding_cuttype and 'CUSTOM' == tbl.jt_prefectbind_cutpos_paper_jobcommand ) then
				return 'TRIPLECUTFINISHSIZECUSTOMLENGTH'
			elseif( 'cutforeedgespecifyfinishingsize' == tbl.perfectbinding_cuttype ) then
				return 'SINGLECUTFINISHSIZELENGTH'
			else
				return nil
			end
		end,
		values = function(value, tbl)
			if( 'on' ~= tbl.perfectbinding ) then
				return nil
			elseif( ('cut3edgesspecifyfinishingsize' == tbl.perfectbinding_cuttype and 'CUSTOM' == tbl.jt_prefectbind_cutpos_paper_jobcommand) or 'cutforeedgespecifyfinishingsize' == tbl.perfectbinding_cuttype ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	perfectbinding_inputtray =	{
		item = 'PERFECTBINDCOVERTRAY',
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				return INPUTTRAY_TABLE
			else
				return nil
			end
		end
	},


	perfectbinding_cutpos_along_mm = {
		item = function(value, tbl)
			if( 'on' == tbl.perfectbinding and 'cut3edgesspecifyfinishingsize' == tbl.perfectbinding_cuttype ) then
				return 'FINISHSIZEVERTADJUSTMENT'
			else
				return nil
			end
		end,
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding and 'cut3edgesspecifyfinishingsize' == tbl.perfectbinding_cuttype ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	trimming = {
		item	= 'TRIMTRIMMODE',
		values	= {
			on	= 'SINGLECUT',
			off	= 'OFF'
		},
		default	= { 'TRIMTRIMMODE', 'OFF' }
	},

	trimming_cutedge_mm = {
		item = 'TRIMSIDEEDGECUT',
		values = function(value, tbl)
			if( 'on' == tbl.trimming ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	jt_custommedianame1 = {
		item = 'MEDIA1NAME',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				return '"'.. GetTrimmedString( value, 229 ) ..'"'
			else
				return nil
			end
		end
	},

	jt_custommedianame2 = {
		item = 'MEDIA2NAME',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				return '"'.. GetTrimmedString( value, 229 ) ..'"'
			else
				return nil
			end
		end
	},

		jt_custommedianame3 = {
		item = 'MEDIA3NAME',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				return '"'.. GetTrimmedString( value, 229 ) ..'"'
			else
				return nil
			end
		end
	},

	jt_custommedianame4 = {
		item = 'MEDIA4NAME',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				return '"'.. GetTrimmedString( value, 229 ) ..'"'
			else
				return nil
			end
		end
	},

	jt_custommedianame5 = {
		item = 'MEDIA5NAME',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				return '"'.. GetTrimmedString( value, 229 ) ..'"'
			else
				return nil
			end
		end
	},

	jt_custommedianame6 = {
		item = 'MEDIA6NAME',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				return '"'.. GetTrimmedString( value, 229 ) ..'"'
			else
				return nil
			end
		end
	},

	jt_custommedianame7 = {
		item = 'MEDIA7NAME',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				return '"'.. GetTrimmedString( value, 229 ) ..'"'
			else
				return nil
			end
		end
	},

	jt_custommedianame8 = {
		item = 'MEDIA8NAME',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				return '"'.. GetTrimmedString( value, 229 ) ..'"'
			else
				return nil
			end
		end
	},

	jt_custommediacharset1 = {
		item = 'MEDIA1NAMECHARSET',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				if( isMember( value, { '1', '2', '3', '5', '6', '8', '9', '13' } )) then
					return value
				else
					return '2'
				end
			else
				return nil
			end
		end
	},

	jt_custommediacharset2 = {
		item = 'MEDIA2NAMECHARSET',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				if( isMember( value, { '1', '2', '3', '5', '6', '8', '9', '13' } )) then
					return value
				else
					return '2'
				end
			else
				return nil
			end
		end
	},

	jt_custommediacharset3 = {
		item = 'MEDIA3NAMECHARSET',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				if( isMember( value, { '1', '2', '3', '5', '6', '8', '9', '13' } )) then
					return value
				else
					return '2'
				end
			else
				return nil
			end
		end
	},

	jt_custommediacharset4 = {
		item = 'MEDIA4NAMECHARSET',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				if( isMember( value, { '1', '2', '3', '5', '6', '8', '9', '13' } )) then
					return value
				else
					return '2'
				end
			else
				return nil
			end
		end
	},

	jt_custommediacharset5 = {
		item = 'MEDIA5NAMECHARSET',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				if( isMember( value, { '1', '2', '3', '5', '6', '8', '9', '13' } )) then
					return value
				else
					return '2'
				end
			else
				return nil
			end
		end
	},

	jt_custommediacharset6 = {
		item = 'MEDIA6NAMECHARSET',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				if( isMember( value, { '1', '2', '3', '5', '6', '8', '9', '13' } )) then
					return value
				else
					return '2'
				end
			else
				return nil
			end
		end
	},

	jt_custommediacharset7 = {
		item = 'MEDIA7NAMECHARSET',
		values = function(value, tbl)
			if( isCustomMedia( tbl ) ) then
				if( isMember( value, { '1', '2', '3', '5', '6', '8', '9', '13' } )) then
					return value
				else
					return '2'
				end
			else
				return nil
			end
		end
	},

	jt_custommediacharset8 = {
		item = function(value, tbl)
			if( true == isCustomMedia( tbl ) ) then
				return 'MEDIA8NAMECHARSET'
			else
				return nil
			end
		end,
		values = function(value, tbl)
			if( true == isCustomMedia( tbl ) ) then
				if(true == isMember( value, { '1', '2', '3', '5', '6', '8', '9', '13' } )) then
					return value
				else
					return '2'
				end
			else
				return nil
			end
		end
	},

	securityprint = {
		item = 'MARKINGTYPE',
		values = function( value, tbl )
			if( 'on' ~= tbl.securityprint ) then
				return 'OFF'
			end
		end
	},

	securityprint_markingtype = {
		item = 'MARKINGTYPE',
		values = function( value, tbl )
			if( 'on' == tbl.securityprint ) then
				local retVal = ({
					copyguard			= 'COPYGUARD',
					securitytrace		= 'SECURITYTRACE',
					alterationdetection	= 'FALSIFICATIONDETECTION'
				})[value]
				return retVal and retVal or 'COPYGUARD'
			end
		end
	},

	copyguard_type = {
		item = { 'UCPPATTERN', 'COPYGUARD', 'COPYGUARDEX', 'SECURITYTRACE', 'PATTERNAREA' },
		values = function( value, tbl )
			if( 'on' == tbl.securityprint and 'copyguard' == tbl.securityprint_markingtype) then
				local retVal = ({
					text								= { 'ON'	, 'OFF'	, 'OFF'	, 'OFF'	, 'FULL'	},
					copyguard							= { 'OFF'	, 'OFF' , 'ON'	, 'OFF'	, 'FULL'	},
					copyguardframe						= { 'OFF'	, 'OFF' , 'ON'	, 'OFF'	, 'FRAME'	},
					textandcopyguardframe				= { 'ON'	, 'OFF' , 'ON'	, 'OFF'	, 'FULL'	},
					textandcopyguard					= { 'ON'	, 'ON'	, 'OFF'	, 'OFF'	, 'FULL'	},
					copyguardandsecuritytrace			= { 'OFF'	, 'OFF' , 'ON'	, 'ON'	, 'FULL'	},
					copyguardandsecuritytraceframe		= { 'OFF'	, 'OFF' , 'ON'	, 'ON'	, 'FRAME'	},
					textandcopyguardandsecuritytrace	= { 'ON'	, 'OFF' , 'ON'	, 'ON'	, 'FULL'	}
				})[value]
				return retVal and retVal or { 'ON'	, 'OFF'	, 'OFF'	, 'OFF'	, 'FULL'	}
			end
		end
	},

	copyguard_color = {
		item = 'UCPPATTERNCOLOR',
		values = function( value, tbl )
			if( 'on' == tbl.securityprint and 'copyguard' == tbl.securityprint_markingtype) then
				local retVal = ({
					black	= 'BLACK',
					cyan	= 'CYAN',
					magenta	= 'MAGENTA'
				})[value]
				return retVal and retVal or 'BLACK'
			end
		end
	},

	copyguard_density = {
		item = 'UCPPATTERNDENSITY',
		values = function( value, tbl )
			if( 'on' == tbl.securityprint and 'copyguard' == tbl.securityprint_markingtype) then
				local retVal = ({
					['1'] = '1',
					['2'] = '2',
					['3'] = '3',
					['4'] = '4',
					['5'] = '5'
				})[value]
				return retVal and retVal or '3'
			end
		end
	},

	copyguard_masktype = {
		item = 'UCPMASKTYPE',
		values = function( value, tbl )
			if( 'on' == tbl.securityprint and 'copyguard' == tbl.securityprint_markingtype) then
				local retVal = ({
					none				= 'NONE',
					wavecrest			= 'SEIGAIHA',
					mesh				= 'AMIME',
					lattice1			= 'KOUSHI1',
					lattice2			= 'KOUSHI2',
					interlockingcircles	= 'SHIPPOU',
					shokkoh				= 'SYOKKOU',
					matsukawabishi		= 'MATSUKAWABISHI',
					scale				= 'UROKO',
					higaki				= 'HIGAKI',
					hexagonal			= 'KIKKOU'
				})[value]
				return retVal and retVal or 'NONE'
			end
		end
	},

	copyguard_effect = {
		item = { 'UCPFGEFFECT', 'UCPBGEFFECT' },
		values = function( value, tbl )
			if( 'on' == tbl.securityprint and 'copyguard' == tbl.securityprint_markingtype) then
				local retVal = ({
					removebgcolor					= { 'REMAIN'	, 'VANISH'	},
					removetextcolor					= { 'VANISH'	, 'REMAIN'	},
					orgnotextcolor					= { 'WHITE'		, 'REMAIN'	},
					orgnotextcolorandremovebgcolor	= { 'WHITE'		, 'VANISH'	},
					orgnobgcolor					= { 'REMAIN'	, 'WHITE'	}
				})[value]
				return retVal and retVal or { 'REMAIN'	, 'VANISH'	}
			end
		end
	},

	copyguard_restrictiontype = {
		item = 'DEVICECONTROLTYPE',
		values = function( value, tbl )
			if( 'on' == tbl.securityprint and 'copyguard' == tbl.securityprint_markingtype) then
				local retVal = ({
					['1'] = '1',
					['2'] = '2',
					['3'] = '3',
					['4'] = '4',
					['5'] = '5',
					['6'] = '6'
				})[value]
				return retVal and retVal or '1'
			end
		end
	},

	securitytrace_embeddedinfo = {
		item = 'TRACEINFO',
		values = function( value, tbl )
			if( 'on' == tbl.securityprint and isMember( tbl.securityprint_markingtype, { 'securitytrace', 'copyguard' } ) ) then
				local retVal = ({
					loginusername	= 'USERID',
					paperid			= 'PAPERID'
				})[value]
				return retVal and retVal or 'USERID'
			end
		end
	},

	securitytrace_embedmode = {
		item = 'EMBEDMODE',
		values = function( value, tbl )
			if( 'on' == tbl.securityprint and 'securitytrace' == tbl.securityprint_markingtype) then
				local retVal = ({
					image		= 'IMAGEQUALITY',
					normal		= 'NORMAL',
					tolerance	= 'TOLERANCE'
				})[value]
				return retVal and retVal or 'NORMAL'
			end
		end
	},

	alterationdetection_density = {
		item = 'PATTERNDENSITY',
		values = function( value, tbl )
			if( 'on' == tbl.securityprint and 'alterationdetection' == tbl.securityprint_markingtype) then
				local retVal = ({
					light	= 'LIGHT',
					middle	= 'MIDDLE',
					dark	= 'DARK'
				})[value]
				return retVal and retVal or 'MIDDLE'
			end
		end
	},

	alterationdetection_errcorrectionlv = {
		item = 'ERRORCORRECTIONLEVEL',
		values = function( value, tbl )
			if( 'on' == tbl.securityprint and 'alterationdetection' == tbl.securityprint_markingtype) then
				local retVal = ({
					['0'] = '0',
					['1'] = '1',
					['2'] = '2',
					['3'] = '3',
					['4'] = '4'
				})[value]
				return retVal and retVal or '4'
			end
		end
	},

	hostmachinename = {
		item = 'HOSTNAME',
		values = function( value, tbl )
			return '"' ..  GetTrimmedString( value, 63 )  .. '"'
		end
	},

	portname = {
		item = 'HOSTPORTNAME',
		values = function( value, tbl )
			return '"' ..  GetTrimmedString( value, 64 )  .. '"'
		end
	},

	printericonname = {
		item = 'HOSTPRINTERNAME',
		values = function( value, tbl )
			return '"' ..  GetTrimmedString( value, 64 )  .. '"'
		end
	}
}

-----------------------------------------------------------------------------
-- ELocal Function : Generate COMMENT
-----------------------------------------------------------------------------
function generate_pcl_comment(settings_table)

	local fields = {
		'RPJL',
		'WIN2000',
		(function()
			if( 'color' == settings_table.colorspec ) then
				return 'PCL6C'
			else
				return 'PCL6'
			end
		end)(),
		(function()
			if( nil	~= settings_table.driverproductversion and
				''	~= settings_table.driverproductversion		) then
				return settings_table.driverproductversion
			else
				return '1.0.0.0'
			end
		end)(),
		(function()
			if( 'color' == settings_table.colorbw ) then
				return 'COLOR'
			else
				return 'GRAYSCALE'
			end
		end)()
	}

	return '@PJL COMMENT '..fields[1]..','..fields[2]..','..fields[3]..','..fields[4]..','..fields[5]..'\r\n'
end
-----------------------------------------------------------------------------
-- ELocal Function : Generate COMMENT OTHER
-----------------------------------------------------------------------------
function generate_pcl_comment_other(settings_table)
	if( nil ~= settings_table.pclother ) then
		return '@PJL COMMENT OTHER = "' ..settings_table.pclother..'"\r\n'
	else
		return nil
	end
end
-----------------------------------------------------------------------------
-- ELocal Function : Generate BEGIN JOB NAME
-----------------------------------------------------------------------------
function generate_begin_jobname(settings_table)
	return '@PJL JOB NAME = ' .. GetJob_Name( settings_table ) .. '\r\n'
end

-----------------------------------------------------------------------------
-- ELocal Function : Generate SCALING in PRINTINFO
-----------------------------------------------------------------------------
function generate_scaling_printinfo(settings_table)
	local scaling_value
	if( 'zoom' == settings_table.reduceenlarge and
		nil ~= settings_table.zoomvalue	 ) then

		if( 50 > tonumber(settings_table.zoomvalue)			) then
			scaling_value = '1'
		elseif( 50 	<= 	tonumber(settings_table.zoomvalue) 	and
				100 > 	tonumber(settings_table.zoomvalue) 		) then
			scaling_value = '2'
		elseif(	100 == 	tonumber(settings_table.zoomvalue)		) then
			scaling_value = '3'
		elseif( 100 < 	tonumber(settings_table.zoomvalue) 	and
				200 >= 	tonumber(settings_table.zoomvalue) 		) then
			scaling_value = '4'
		elseif(	200 <	tonumber(settings_table.zoomvalue)		) then
			scaling_value = '5'
		else
			scaling_value = '3'
		end

	elseif( 'fittoprintsize' == settings_table.reduceenlarge 	and
			nil ~= settings_table.jt_print_paperwidth_mm		and
			nil ~= settings_table.jt_print_paperheight_mm		and
			nil ~= settings_table.jt_document_paperwidth_mm		and
			nil ~= settings_table.jt_document_paperheight_mm		) then
		local scaling = 1.0

		local x_scaling = settings_table.jt_print_paperwidth_mm / settings_table.jt_document_paperwidth_mm
		local y_scaling = settings_table.jt_print_paperheight_mm  / settings_table.jt_document_paperheight_mm

		if( x_scaling < y_scaling ) then
			scaling = x_scaling
		elseif( y_scaling < x_scaling)  then
			scaling = y_scaling
		elseif( x_scaling == y_scaling ) then
			scaling = x_scaling
		else
			scaling = 0.1
		end

		if( 	0.5 > 	scaling 		) then
			scaling_value = '1'
		elseif( 0.5 <= 	scaling	and
				1.0 >	scaling 		) then
			scaling_value = '2'
		elseif( scaling == 1.0 			) then
			scaling_value = '3'
		elseif( 1.0 < 	scaling 	and
				2.0 >= 	scaling 		) then
			scaling_value = '4'
		elseif( 2.0 <	scaling			) then
			scaling_value = '5'
		else
			scaling_value = '3'
		end
	else
		scaling_value = '3'
	end
	return 'SCALING='..scaling_value
end
-----------------------------------------------------------------------------
-- ELocal Function : Generate DULPEX_NUP in PRINTINFO
-----------------------------------------------------------------------------
function generate_duplex_nup_printinfo(settings_table)
	local duplex_nup_value

	-- 1 mojime
	duplex_nup_value = '0'
	-- 2 mojime
	duplex_nup_value = duplex_nup_value ..'0'
	-- 3 mojime
	if( '9pagespersheet' == settings_table.layout ) then
		duplex_nup_value = duplex_nup_value ..'1'
	elseif( '16pagespersheet' == settings_table.layout ) then
		duplex_nup_value = duplex_nup_value ..'2'
	else
		duplex_nup_value = duplex_nup_value ..'0'
	end
	-- 4 mojime
	if( '2pagespersheet' == settings_table.layout ) then
		duplex_nup_value = duplex_nup_value ..'1'
	elseif( '4pagespersheet' == settings_table.layout ) then
		duplex_nup_value = duplex_nup_value ..'2'
	elseif( '6pagespersheet' == settings_table.layout ) then
		duplex_nup_value = duplex_nup_value ..'4'
	else
		duplex_nup_value = duplex_nup_value ..'0'
	end
	-- 5 mojime
	duplex_nup_value = duplex_nup_value ..'0'
	-- 6 mojime
	if( (settings_table.booklet)	and
		'off' ~= settings_table.booklet	) then
		duplex_nup_value = duplex_nup_value ..'8'
	elseif(	(settings_table.layout)		and
			'1pagepersheet' ~= settings_table.layout	) then
		if( (settings_table.duplex)	and
			'off' ~= settings_table.duplex ) then
			duplex_nup_value = duplex_nup_value ..'2'
		else
			duplex_nup_value = duplex_nup_value ..'1'
		end
	else
		duplex_nup_value = duplex_nup_value ..'0'
	end
	-- 7 mojime
	duplex_nup_value = duplex_nup_value ..'0'
	-- 8 mojime
	if(	'1pagepersheet' == settings_table.layout 	and
		(settings_table.duplex)						and
		'off' ~= settings_table.duplex					) then
		duplex_nup_value = duplex_nup_value..'1'
	else
		duplex_nup_value = duplex_nup_value..'0'
	end

	return 'DUPLEX_NUP='..duplex_nup_value
end
-----------------------------------------------------------------------------
-- ELocal Function : Generate OTHER in PRINTINFO
-----------------------------------------------------------------------------
function generate_other_printinfo(settings_table)
	local othe_value = 0

	if('on' == settings_table.wmk_print 		) then
		othe_value = othe_value + 8
	end

	if(	nil ~= settings_table.booklet	and
		'off' ~= settings_table.booklet 		) then
		othe_value = othe_value + 4
	end

	if(	'on' == settings_table.frontcover 	or
		'on' == settings_table.backcover 	or
		'on' == settings_table.slipsheet 	or
		'on' == settings_table.chapterpages 	) then
		othe_value = othe_value + 2
	end

	return string.format('OTHER=0000000%X', othe_value)

end
-----------------------------------------------------------------------------
-- ELocal Function : Generate PRINTINFO
-----------------------------------------------------------------------------
function generate_printinfo(settings_table)
	local printinfo = '@PJL SET PRINTINFO = "'
	printinfo = printinfo..generate_duplex_nup_printinfo(settings_table)..'/'
	printinfo = printinfo..generate_scaling_printinfo(settings_table)..'/'
	printinfo = printinfo..generate_other_printinfo(settings_table)..'"\r\n'

	return printinfo
end

-----------------------------------------------------------------------------
-- ELocal Function : Generate PROOFJOB
-----------------------------------------------------------------------------
function generate_jobtype(settings_table)
	if( isInvalidJobCode( settings_table ) ) then
		return nil
	else
		return ({
			sampleprint				= '@PJL PROOFJOB\r\n',
			lockedprint				= '@PJL SECUREJOB\r\n',
			holdprint				= '@PJL PAUSEDJOB\r\n',
			storedprint				= '@PJL PAUSEDJOB\r\n',
			storedprintshared		= '@PJL PAUSEDJOB\r\n',
			documentserverrgb		= '@PJL REUSABLESTOREDJOB\r\n'
		})[settings_table.jobtype]
	end
end

-----------------------------------------------------------------------------
-- Export Function : Generate Command
-----------------------------------------------------------------------------
function generate_command(settings_table)

  local converter = TableConverter(convert_table)
  local job_commands = converter:run(settings_table)

  local commands = {}

  table.insert( commands, string.char(0x1b) .. '%-12345X')
  table.insert( commands, generate_begin_jobname(settings_table))
  table.insert(	commands, generate_printinfo(settings_table) )
  table.insert(	commands, generate_pcl_comment(settings_table) )
  table.insert(	commands, generate_pcl_comment_other(settings_table) )
  table.insert(	commands, generate_jobtype(settings_table) )


  for item, value in pairs(job_commands) do
	table.insert(commands, string.format('@PJL SET %s = %s\r\n', item, value))
  end

  local loader = ExtensionLoader(_MODULE, 'generate_command' )
  for name, func in loader:extensions() do
     func(settings_table, commands)
  end

  table.insert( commands, '@PJL ENTER LANGUAGE = PCLXL\r\n' )

  return table.concat(commands)
end

return true
